
import * as fs from 'fs/promises';

import { Node } from './node-types/abstract-node';
import { QuestionNode } from "./node-types/question";
import { ClaimNode } from './node-types/claim';
import { EvidenceNode } from './node-types/evidence';
import { SourceNode } from './node-types/source';

import { Edge } from './edge-types/abstract-edge';
import { GenerateEdge } from './edge-types/generate';
import { InformEdge } from './edge-types/inform';
import { SupportEdge } from './edge-types/support';
import { OpposeEdge } from './edge-types/oppose';

export class Compiler {

    public static nodes = [
        ClaimNode,
        EvidenceNode,
        QuestionNode,
        SourceNode
    ];

    public static edges = [
        GenerateEdge,
        InformEdge,
        SupportEdge,
        OpposeEdge
    ];

    public static getNodeType(name: string) {
        for (const nodeType of Compiler.nodes) {
            if (nodeType.TYPE_PATTERN.test(name)) return nodeType;
        }
        throw new Error(`File name does not match any known node type \n - ${name}`);
    }

    public static getEdgeType(name: string) {
        for (const edgeType of Compiler.edges) {
            if (name.includes(edgeType.EDGE_TYPE)) return edgeType;
        }
        throw new Error(`Edge name does not match any known edge types \n - ${name}`);
    }

    public static async loadNodes(directory: string) {
        console.log('\nLoading nodes...');
        const nodes = [];
        const files = await fs.readdir(directory);
        for (const file of files) try {
            const data = await fs.readFile(`${directory}/${file}`, 'utf-8');
            const NodeType = Compiler.getNodeType(file);
            const node = NodeType.load(file, data);
            nodes.push(node);
            console.log(`Loaded ${NodeType.TYPE} node "${node.name}"`);
        } catch (err) {
            console.error(err.message);
        }
        return nodes;
    }

    public static loadEdges(nodes: Node[]) {
        console.log('\nLoading edges...');
        const edges = [];
        for (const node of nodes) {
            const references = node.parseDataEdges();
            for (const reference of references) try {
                const source = node;
                const destination = nodes.find(node => node.name === reference.referenceName);
                if (!destination) throw new Error(`Could not find referenced node "${reference.referenceName}"`);
                const EdgeClass = Compiler.getEdgeType(reference.edgeType);
                const edge = EdgeClass.createEdge(source, destination);
                edges.push(edge);
                console.log(`Loaded ${EdgeClass.EDGE_TYPE} edge from "${edge.source.name}" to "${edge.reference.name}"`);
            } catch (err) {
                console.error(err.message);
            }
        }
        return edges;
    }

    public static registerEdges(edges: Edge[]) {
        console.log('\nRegistering edges...');
        for (const edge of edges) try {
            edge.source.registerEdge(edge);
            edge.reference.registerEdge(edge);
            console.log(`Registered ${edge.constructor.name} from "${edge.source.name}" to "${edge.reference.name}"`);
        } catch (err) {
            console.error(err.message);
        }
    }

    public static traverseGraph(node: Node, relationship: string | null, indent: number = 0) {
        const currentIndent = ' '.repeat(indent);
        const nodeType = node.constructor as typeof Node;
        const bullet = relationship ? '> ' + relationship : '-';
        console.log(`${currentIndent}${bullet} ${nodeType.SYMBOL} ${node.name}`);
        for (const edge of node.incomingReferences) {
            if (edge.source instanceof QuestionNode) continue;
            const edgeType = edge.constructor as typeof Edge;
            Compiler.traverseGraph(edge.source, edgeType.RELATIONSHIP, indent + 2);
        }
    }

}

async function runTest() {
    console.log("Running Test");
    const graphPath = "./example_graph";
    const nodes = await Compiler.loadNodes(graphPath);
    const edges = await Compiler.loadEdges(nodes);
    Compiler.registerEdges(edges);
    console.log("\nGenerating Artifact...");
    const questions = nodes.filter(node => node instanceof QuestionNode);
    for (const question of questions) Compiler.traverseGraph(question, null);
    console.log("\nTest Complete");
}


runTest().then(() => console.log("Done"));